import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.Buffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Timer;

import mensajesSIP.ACKMessage;
import mensajesSIP.BusyHereMessage;
import mensajesSIP.ByeMessage;
import mensajesSIP.InviteMessage;
import mensajesSIP.NotFoundMessage;
import mensajesSIP.OKMessage;
import mensajesSIP.RegisterMessage;
import mensajesSIP.RequestTimeoutMessage;
import mensajesSIP.RingingMessage;
import mensajesSIP.SDPMessage;
import mensajesSIP.SIPException;
import mensajesSIP.SIPMessage;
import mensajesSIP.ServiceUnavailableMessage;
import mensajesSIP.TryingMessage;


public class Proxy {

	private static boolean looserouting,debug;
	private static DatagramSocket socket;
	private static int puerto;
	private static DatagramPacket mensaje;
	private static final int MAXFORWARD=70;
	public static final String dominio="dominio";
	private static HashMap<String,UsuarioRegistrado>listaUsuariosRegistrados=new HashMap<String,UsuarioRegistrado>();
	private static HashMap<String,Llamada>listallamadas=new HashMap<String,Llamada>();
	private static final int ESTADO_LIBRE=0;
	private static final int ESTADO_PROCESANDO_INVITE=1;
	private static int estado=ESTADO_LIBRE;
	private static HashMap<String,String>listaAutorizados=new HashMap<String,String>();
	private static final String RUTA="autorizados.txt";

	public static void main(String[] args){
		if(Proxy.parsearParametros(args)){
			cargarlista();
			try {
				socket=new DatagramSocket(puerto);//pongo el servidor a escuchar en un determinado puerto
				//System.out.println("funciona");
			} catch (SocketException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			//System.out.println("conectando");
			ejecutarServidor();
		}else{
			System.out.println("Error en los parametros");
			System.out.println("Parametros: puerto loose-routing debug");
			System.out.println("Sin parametros comportamiento por defecto");
		}
		
		
		
	}

	private static void cargarlista() {
		File fichero=new File(RUTA);
		try {
			FileInputStream fis=new FileInputStream(fichero);
			InputStreamReader isr=new InputStreamReader(fis);
			BufferedReader br=new BufferedReader(isr);
			String nombre="";
			while((nombre=br.readLine())!=null){
				listaAutorizados.put(nombre, null);//he usado hasmap xk tiene acceso directo por clave
				//al nombre. Mientras que para buscar en un arraylist necesito recorrerlo.
			}
			br.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		
	}

	private static boolean parsearParametros(String[] args) {
		switch (args.length) {

		case 0:
			puerto = 5060;
			looserouting = true;
			debug = true;
			break;
		case 3:
			puerto = Integer.parseInt(args[0]);
			looserouting = new Boolean(args[1]);
			debug = new Boolean(args[2]);
			break;
		default:
			return false;
		}
		return true;
	}

	private static void ejecutarServidor() {
		// TODO Auto-generated method stub
		while(true){
			byte[] datosRecibidos=new byte[1024];
			SIPMessage mensajeSip;
			
			DatagramPacket paqueteRecibido = new DatagramPacket(datosRecibidos, datosRecibidos.length);
		      try {
				System.out.print("esperando mensaje...\t");
		    	  socket.receive(paqueteRecibido);
				System.out.println("mensaje Recibido");
				mensajeSip=SIPMessage.parseMessage(new String(paqueteRecibido.getData()));
				procesar(mensajeSip);
				
			} catch (IOException e) {
				System.out.println("Error en el socket");
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SIPException e) {
				// TODO Auto-generated catch block
				System.out.println("Error en la creacion del sipmessage");
				e.printStackTrace();
			}
		}
	}

	private static void procesar(SIPMessage mensajeSip) {
	//el proxy aqui procesa los mensajes que le llegan
		if(mensajeSip instanceof RegisterMessage){
			procesarMensajeRegistro((RegisterMessage)mensajeSip); 
			
		}else if(mensajeSip instanceof InviteMessage){
			procesarMensajeInvite((InviteMessage)mensajeSip); 	
			 
		}else if(mensajeSip instanceof RingingMessage){
			procesarMensajeRinging((RingingMessage)mensajeSip);
		}else if(mensajeSip instanceof OKMessage){
			procesarMensajeOK((OKMessage)mensajeSip);
		}else if(mensajeSip instanceof BusyHereMessage){
			procesarMensajeBusy((BusyHereMessage)mensajeSip);
		}else if(mensajeSip instanceof ACKMessage){
			procesarMensajeACK((ACKMessage)mensajeSip);
		}else if(mensajeSip instanceof ByeMessage){
			procesarMensajeBye((ByeMessage)mensajeSip);
		}
		else{
				System.out.println("Mensaje aun no implementado");
			}	
		
	}
	private static void procesarMensajeBye(ByeMessage mensajeSip) {
		System.out.println("\t Recibido Bye desde: "+ mensajeSip.getFromName());
		
		UsuarioRegistrado u=listaUsuariosRegistrados.get(mensajeSip.getToName());
		if(u!=null){
			mensajeSip.addVia(getMiVia());
			enviarMensaje(mensajeSip, u.getIp(), u.getPuerto());
			System.out.println("Bye reenviado");
			Proxy.estado=ESTADO_LIBRE;
			Extremos extremos=new Extremos(mensajeSip.getToName(),mensajeSip.getFromName());
			//listallamadas.remove(extremos.getLlamante()).setEstado(Llamada.HABLANDO);
			listaUsuariosRegistrados.get(extremos.getLlamante()).setComunica(false);
			listaUsuariosRegistrados.get(extremos.getLlamado()).setComunica(false);
		}else{
			System.out.println("Error reenviando Bye");
		}
	}

	private static void procesarMensajeACK(ACKMessage mensajeSip) {
		if(mensajeSip.getcSeqStr().contains("ACK")){
			System.out.println("\t recibido ACK desde: "+mensajeSip.getFromName());
			if(listallamadas.get(mensajeSip.getFromName())!=null){//si la llamada se ha establecido es que existe y hay que reenviar ese ACK 
											// en cualquier otro caso no hay que reenviarlo
				reenviarACK(mensajeSip);
			}
			
		}else if(mensajeSip.getcSeqStr().contains("REGISTER")){
			System.out.println("\t recibido ACK de REGISTER desde: "+mensajeSip.getFromName());
		}
		
	
	}

	private static void reenviarACK(ACKMessage mensajeSip) {
		
		UsuarioRegistrado u=listaUsuariosRegistrados.get(mensajeSip.getToName());
		if(u!=null){
			mensajeSip.addVia(getMiVia());
			enviarMensaje(mensajeSip, u.getIp(), u.getPuerto());
			System.out.println("ACK reenviado");
		}else{
			System.out.println("Error reenviando ACK");
		}
	}

	private static void procesarMensajeBusy(BusyHereMessage mensajeSip) {
		System.out.println("\t recibido Busy Here(486) No ha descolgado: "+mensajeSip.getToName());
		mensajeSip.deleteVia();
		UsuarioRegistrado u=listaUsuariosRegistrados.get(mensajeSip.getFromName());
		enviarMensaje(mensajeSip, u.getIp(), u.getPuerto());
	}

	private static void procesarMensajeOK(OKMessage mensajeSip) {
		Extremos extremos=new Extremos(mensajeSip.getToName(),mensajeSip.getFromName());
		listallamadas.get(extremos.getLlamante()).setEstado(Llamada.HABLANDO);
		listaUsuariosRegistrados.get(extremos.getLlamante()).setComunica(true);
		listaUsuariosRegistrados.get(extremos.getLlamado()).setComunica(true);
		System.out.println("\t recibido han descolgado(200OK): "+mensajeSip.getToName());
		
		//enviamos 200ok ahora al usuario
		UsuarioRegistrado u= listaUsuariosRegistrados.get(extremos.getLlamante());
		quitarMivia(mensajeSip);
		
		enviarMensaje(mensajeSip,u.getIp(),u.getPuerto());
	}

	private static void quitarMivia(OKMessage mensajeSip) {
		ArrayList<String> vias=mensajeSip.getVias();
		vias.remove(vias.size()-1);
	}

	private static void procesarMensajeRinging(RingingMessage mensajeSip) {
		ArrayList<String>vias=mensajeSip.getVias();
		vias.remove(vias.size()-1);
		UsuarioRegistrado u = listaUsuariosRegistrados.get(mensajeSip.getFromName());

		System.out.println("\tRecibido sonando (180Ringring) en: " + mensajeSip.getToName());
		
		if (u == null) {
			System.out.println("error no se encontro el usuario: "
					+ mensajeSip.getFromName());
		} else if (!enviarMensaje(mensajeSip, u.getIp(), u.getPuerto())) {
			System.out.println("\tInvite: Error al enviar mensaje 180Ringing");
		}
	}

	private static void procesarMensajeInvite(InviteMessage mensajeInvite) {
		if (estado == ESTADO_LIBRE) {
			
			String nombre = mensajeInvite.getDestination()
					.replaceFirst("sip:", "").replaceFirst("@" + dominio, "");
			UsuarioRegistrado u = listaUsuariosRegistrados.get(nombre);// busco en la lista
															// de usuarios el
															// usuario al que se
															// invita
			if (u == null) {
				System.out.println("\tInvite: Error usuario"
						+ mensajeInvite.getDestination() + "no existe");
				enviar404(mensajeInvite);
			} else {

				if (u.getComunica()) {
					System.out.println("\tInvite: Error usario"
							+ mensajeInvite.getDestination() + "comunica");
					// ENVIAR MENSAJE 486
				} else {
					
					estado = ESTADO_PROCESANDO_INVITE;
					Extremos e= new Extremos(mensajeInvite.getToName(),mensajeInvite.getFromName());
					listallamadas.put(e.getLlamante(), new Llamada(e));
					System.out
							.println("\tInvite: No comunica, llamando (enviando invite): "
									+ mensajeInvite.getToName());
					// enviar mensaje 100 intentadolo
					String nombreorigen = mensajeInvite.getFromName();
					UsuarioRegistrado uorigen = listaUsuariosRegistrados.get(nombreorigen);
					enviar100(mensajeInvite, uorigen.getIp(),
							uorigen.getPuerto());

					/*
					 * if(looserouting){
					 * System.out.println("Poner record-route"); }
					 */
					// crear sdp que no nos parsea, es decir crear sdp de nuevo porque manteniendo el anterior da error
					SDPMessage sdp = new SDPMessage();
					ArrayList<Integer> opcionesSdp = new ArrayList<Integer>();
					opcionesSdp.add(96);
					opcionesSdp.add(97);
					opcionesSdp.add(98);
					sdp.setOptions(opcionesSdp);// preguntar xk no parsea el sdp
					
					sdp.setIp(uorigen.getIp());
					sdp.setPort(uorigen.getPuerto());
					mensajeInvite.setSdp(sdp);

					mensajeInvite.addVia(getMiVia());// GENERAR
																	// LA VIA
																	// DEL PROXY

					// Crear ueva llamada
					Extremos extremos = new Extremos(
							mensajeInvite.getToName(),
							mensajeInvite.getFromName());
					Llamada llamada = new Llamada(extremos);
					listallamadas.put(e.getLlamante(), llamada);

					// Configurar tarea del timer
					TareaInvite tarea = new TareaInvite(llamada);
					Timer timer = new Timer();// el timer usa la tarea al
												// ejecutarse

					enviarMensaje(mensajeInvite, u.getIp(), u.getPuerto());
					// activar el timer
					timer.schedule(tarea, 10000);// para que si a los 10
													// segundos no a respondido
													// ejecute la tarea que
													// enviara el 408

				}

			}
		}else{
			enviar503(mensajeInvite);
		}

	}

	private static void enviar404(InviteMessage mensajeSip) {
		NotFoundMessage mensaje404=new NotFoundMessage();
		mensaje404.setVias(mensajeSip.getVias());
		mensaje404.setToUri(mensajeSip.getToUri());
		mensaje404.setToName(mensajeSip.getToName());
		mensaje404.setFromUri(mensajeSip.getFromUri());
		mensaje404.setFromName(mensajeSip.getFromName());
		mensaje404.setcSeqNumber("2 INVITE\n");
		mensaje404.setContentLength(0);
		mensaje404.setCallId(mensajeSip.getCallId());
		
		String nombre = mensajeSip.getFromName();
		
		UsuarioRegistrado u = listaUsuariosRegistrados.get(nombre);
		
		enviarMensaje(mensaje404, u.getIp(), u.getPuerto());
		
		
	}

	private static void enviar503(InviteMessage mensajeSip) {
		ServiceUnavailableMessage mensaje503=new ServiceUnavailableMessage();
		mensaje503.setVias(mensajeSip.getVias());
		mensaje503.setToUri(mensajeSip.getToUri());
		mensaje503.setToName(mensajeSip.getToName());
		mensaje503.setFromUri(mensajeSip.getFromUri());
		mensaje503.setFromName(mensajeSip.getFromName());
		mensaje503.setcSeqNumber("2 INVITE\n");
		mensaje503.setContentLength(0);
		mensaje503.setCallId(mensajeSip.getCallId());
		String nombre = mensajeSip.getFromName();
				
		UsuarioRegistrado u = listaUsuariosRegistrados.get(nombre);// busco en la lista
														// de usuarios el
														// usuario al que se
														// invita
		
		enviarMensaje(mensaje503, u.getIp(), u.getPuerto());
	}

	private static void enviar100(InviteMessage mensajeSip, String ip, int puerto) {
		TryingMessage mensaje100=new TryingMessage();
		mensaje100.setVias(mensajeSip.getVias());
		mensaje100.setToUri(mensajeSip.getToUri());
		mensaje100.setToName(mensajeSip.getToName());
		mensaje100.setFromUri(mensajeSip.getFromUri());
		mensaje100.setFromName(mensajeSip.getFromName());
		mensaje100.setcSeqNumber("2 INVITE\n");
		mensaje100.setContentLength(0);
		mensaje100.setCallId(mensajeSip.getCallId());

		if (!enviarMensaje(mensaje100, ip, puerto))
			System.out.println("\n Error en el envio de mensaje100");
	}

	private static void procesarMensajeRegistro(RegisterMessage mensajeSip) {
		
		if (estaAutorizado(mensajeSip.getFromName())) {

			UsuarioRegistrado u = new UsuarioRegistrado(mensajeSip);//creamos usuario a partir del mensaje register
			listaUsuariosRegistrados.put(u.getNombre(), u);//metemos ese usuario en la lista de usuarios registrados
			boolean ok = responderOk(u.getIp(), u.getPuerto(),
					u.getUsuarioSip(), (mensajeSip).getVias());
			if (ok) {
				System.out.println("\tRegister de: "
						+ (mensajeSip).getFromName());
			} else {
				System.out.println("\tError enviando ok del registro");
			}
		} else {
			enviar404(mensajeSip);
		}
	}

	private static void enviar404(RegisterMessage mensajeSip) {
		NotFoundMessage mensaje404=new NotFoundMessage();
		mensaje404.setVias(mensajeSip.getVias());
		mensaje404.setToUri(mensajeSip.getToUri());
		mensaje404.setToName(mensajeSip.getToName());
		mensaje404.setFromUri(mensajeSip.getFromUri());
		mensaje404.setFromName(mensajeSip.getFromName());
		mensaje404.setcSeqNumber("1 REGISTER\n");
		mensaje404.setContentLength(0);
		mensaje404.setCallId(mensajeSip.getCallId());
		
		
		String[] contacto=mensajeSip.getContact().split(":");
		String ip=contacto[0];
		int puerto=Integer.parseInt(contacto[1]);
		enviarMensaje(mensaje404, ip,puerto);
		
	}

	private static boolean estaAutorizado(String name) {
		return listaAutorizados.containsKey(name);
	}

	private static String getMiVia(){
		try {
			return InetAddress.getLocalHost().getHostAddress().toString()+":"+puerto+"\n";
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}
	private static boolean responderOk(String ip, int puerto, String usuarioSip, ArrayList<String> vias) {
		
		
		OKMessage mensajeOk;
		byte[] mensajeBytes;
		double numaleatorio=Math.round(Math.random()*1000000);
		
		mensajeOk=new OKMessage();
		
		mensajeOk.setVias(vias );
		mensajeOk.setFromUri(usuarioSip);
		mensajeOk.setToName(usuarioSip.substring(4, usuarioSip.indexOf('@')));
		mensajeOk.setToUri(usuarioSip);
		mensajeOk.setCallId(Double.toString(numaleatorio));
		mensajeOk.setcSeqNumber("1 REGISTER\n");
		mensajeOk.setContact(ip+":"+puerto);
		mensajeOk.setExpires("120");
		mensajeOk.setContentLength(0);
		//una vez creado el mensaje de registro y tras convertirlo en un array de bytes, 
		//metere ese mensaje en el datagramPacket para enviarlo
		return enviarMensaje(mensajeOk,ip,puerto);
		
		
	}
	
	private static boolean enviarMensaje(SIPMessage msg,String ip,int puerto){
		byte[] mensajeBytes=msg.toStringMessage().getBytes();
		
		try {
			DatagramPacket mensaje=new DatagramPacket(mensajeBytes,mensajeBytes.length,InetAddress.getByName(ip),puerto);
			socket.send(mensaje);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public static void envia408(Extremos extremos) {
		
		double numaleatorio=Math.round(Math.random()*1000000);

		RequestTimeoutMessage mensaje408;
		ArrayList<String> vias=new ArrayList<String>();
		vias.add(getMiVia());
		mensaje408=new RequestTimeoutMessage();
		
		mensaje408.setVias(vias );
		
		
		mensaje408.setFromUri("sip:"+extremos.getLlamante()+"@"+dominio);
		mensaje408.setFromName(extremos.getLlamante());
		mensaje408.setToUri("sip:"+extremos.getLlamado()+"@"+dominio);
		mensaje408.setToName(extremos.getLlamado());
		mensaje408.setCallId(Double.toString(numaleatorio));
		mensaje408.setcSeqNumber("2 INVITE\n");
		mensaje408.setContentLength(142);//que longitud del mensaje?
		UsuarioRegistrado u=listaUsuariosRegistrados.get(extremos.getLlamante());
		if (!enviarMensaje(mensaje408,u.getIp(),u.getPuerto())){
			System.out.println("Error en envio de Mensaje 408");
		}
		estado=ESTADO_LIBRE;
	}
}
