import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Scanner;

import mensajesSIP.ACKMessage;
import mensajesSIP.BusyHereMessage;
import mensajesSIP.ByeMessage;
import mensajesSIP.InviteMessage;
import mensajesSIP.NotFoundMessage;
import mensajesSIP.OKMessage;
import mensajesSIP.RegisterMessage;
import mensajesSIP.RequestTimeoutMessage;
import mensajesSIP.RingingMessage;
import mensajesSIP.SDPMessage;
import mensajesSIP.SIPException;
import mensajesSIP.SIPMessage;
import mensajesSIP.ServiceUnavailableMessage;
import mensajesSIP.TryingMessage;


public class HiloEscucha extends Thread{

	private UA cliente;
	private static final String prefijo="\n \t RECIBIDO: ";
	private int estado;
	private static int SIN_ESTADO=0;
	private static int RESPUESTA_ENVIADA=1;
	public HiloEscucha(UA cliente){
		this.cliente=cliente;
		estado=SIN_ESTADO;
	}
	public void run(){
		SIPMessage mensajeSip;
		
		byte[] datosRecibidos=new byte[1024];
		DatagramPacket paqueteRecibido = new DatagramPacket(datosRecibidos, datosRecibidos.length);
		
		while(true){
		      try {
				cliente.getSocket().receive(paqueteRecibido);
				mensajeSip=SIPMessage.parseMessage(new String(paqueteRecibido.getData()));//paso de paquete a mensajesip que es lo que necesito para procesar mi mensaje
				procesar(mensajeSip);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SIPException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		      
		}
	}
	private  void procesar(SIPMessage mensajeSip) {
		if(mensajeSip instanceof InviteMessage){
			if(estado!=RESPUESTA_ENVIADA){
				procesarInvite ((InviteMessage) mensajeSip);
			}
			
		}else if(mensajeSip instanceof RequestTimeoutMessage){
			System.out.println(prefijo+"NO han descolgado, mensaje 408 \n");
			//cliente.getInterprete().despertar();
			//cliente.iniciarInterprete();
			responderACK((RequestTimeoutMessage)mensajeSip);
			
		}else if(mensajeSip instanceof TryingMessage){
			System.out.println(prefijo+"Intentando \n");
		}else if(mensajeSip instanceof RingingMessage){
			System.out.println(prefijo+"Telefono sonando en: "+((RingingMessage)mensajeSip).getToName()+"\n");
		}else if(mensajeSip instanceof ServiceUnavailableMessage){
			System.out.println(prefijo+"Servidor ocupado");
			//cliente.getInterprete().despertar();
			//cliente.iniciarInterprete();
		}else if(mensajeSip instanceof NotFoundMessage){
			procesar404((NotFoundMessage) mensajeSip);
			//cliente.getInterprete().despertar();
			//cliente.iniciarInterprete();
			
		}else if(mensajeSip instanceof OKMessage){
			responderACK((OKMessage)mensajeSip);
			System.out.println(prefijo+"Ha descolgado: "+ ((OKMessage) mensajeSip).getToName());
			//cliente.getInterprete().despertar();
			//cliente.iniciarInterprete();
		}else if(mensajeSip instanceof BusyHereMessage){
			cliente.setUsuarSIPotro(null);
			responderACK((BusyHereMessage)mensajeSip);
			System.out.println(prefijo+"Extremo ocupado o comunica o no ha descolgado: "+ ((BusyHereMessage) mensajeSip).getToName());
			//cliente.getInterprete().despertar();
			//cliente.iniciarInterprete();
		}else if(mensajeSip instanceof ACKMessage){
			System.out.println(prefijo+"Confirmacion recibida");
			estado=SIN_ESTADO;
		}else if(mensajeSip instanceof ByeMessage){
			System.out.println(prefijo+"Bye Recibido. Llamada finalizada");
			//cliente.getInterprete().stop();
			//cliente.terminarInterprete();
			//cliente.iniciarInterprete();
			//cliente.reiniciaScanner();
			cliente.setUsuarSIPotro(null);
		}
		else{
			System.out.println(prefijo+"Mensaje aun no implementado \n");
		}
		
		
		
	}
	
	
	private void responderACK(RequestTimeoutMessage mensajeSip) {
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias() );
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getFromName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("2 INVITE\n");
		ack.setMaxForwards(UA.MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
		
	}
	private void procesar404(NotFoundMessage mensajeSip) {
		System.out.println(prefijo+"Error el usuario no existe");
		cliente.setUsuarSIPotro(null);
		responderACK(mensajeSip);
	}
	private void procesarInvite(InviteMessage mensajeSip) {
		
		enviar180ringing((InviteMessage) mensajeSip);
		cliente.getInterprete().stop();
		
		System.out.println(prefijo+"llamada. Descolgar?(S/N): ");
		
		cliente.reiniciaScanner();
		String c="";
		while(!cliente.getLector().hasNextLine());
		do{
			c = cliente.getLector().nextLine();
			
		}while (!c.equalsIgnoreCase("s") && !c.equalsIgnoreCase("n"));
		estado=RESPUESTA_ENVIADA;
		cliente.iniciarInterprete();
		
		if(c.equalsIgnoreCase("S")){//si descuelga  mandamos el 
			cliente.setUsuarSIPotro(mensajeSip.getFromName());
			enviar200Ok(mensajeSip);
		}else if(c.equalsIgnoreCase("N")){ //si no descuelga enviamos 408
			enviar486(mensajeSip);
		}	
	}
	private void enviar180ringing(InviteMessage mensajeSip) {
	
		//Crear mensaje Ring
		RingingMessage mensaje180 = new RingingMessage();
		mensaje180.setToUri(mensajeSip.getToUri());
		mensaje180.setToName(mensajeSip.getToName());
		mensaje180.setFromUri(mensajeSip.getFromUri());
		mensaje180.setFromName(mensajeSip.getFromName());
		mensaje180.setContentLength(0);
		//mensaje180.setContact(arg0);
		mensaje180.setCallId(mensajeSip.getCallId());
		mensaje180.setcSeqNumber("2 INVITE\n");
		mensaje180.setVias(mensajeSip.getVias());
		//enviar mensaje:
		
		if(!enviarMensaje(mensaje180)){
			System.out.println("Error al enviar mensaje 180 Ringring");
		}
	
		
	}
	private String getMiVia(){
		try {
			return InetAddress.getLocalHost().getHostAddress().toString()+":"+cliente.getPuertoUa()+"\n";
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}
	private boolean enviarMensaje(SIPMessage msg) {
		byte[] mensajeBytes=msg.toStringMessage().getBytes();
		DatagramPacket mensaje=new DatagramPacket(mensajeBytes,mensajeBytes.length,cliente.getIpProxy(),cliente.getPuertoProxy());
		
		try {
			cliente.getSocket().send(mensaje);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} 
		return true;
		
	}
	private void enviar486(InviteMessage mensajeSip) {
		
		BusyHereMessage mensajebusy486=new BusyHereMessage();
		
		mensajebusy486.setVias(mensajeSip.getVias());
		mensajebusy486.setFromUri(mensajeSip.getFromUri());
		mensajebusy486.setToName(mensajeSip.getToName());
		mensajebusy486.setFromName(mensajeSip.getFromName());
		mensajebusy486.setToUri(mensajeSip.getToUri());
		mensajebusy486.setCallId(mensajeSip.getCallId());
		mensajebusy486.setcSeqNumber("2 INVITE\n");
		mensajebusy486.setContentLength(0);
		
		enviarMensaje(mensajebusy486);
		
	}
	private boolean enviar200Ok(InviteMessage mensajeSip) {
		OKMessage mensajeOk;
		
		
		mensajeOk=new OKMessage();
		
		mensajeOk.setVias(mensajeSip.getVias() );
		mensajeOk.setFromUri(mensajeSip.getFromUri());
		mensajeOk.setToName(mensajeSip.getToName());
		mensajeOk.setFromName(mensajeSip.getFromName());
		mensajeOk.setToUri(mensajeSip.getToUri());
		mensajeOk.setCallId(mensajeSip.getCallId());
		mensajeOk.setcSeqNumber("2 INVITE\n");
		mensajeOk.setContact(mensajeSip.getContact());
		mensajeOk.setContentLength(142);//cual es la longitud .misma q invite?
		
		
		SDPMessage sdp=new SDPMessage();
		ArrayList<Integer>opcionesSdp=new ArrayList<Integer>();
		opcionesSdp.add(96);
		opcionesSdp.add(97);
		opcionesSdp.add(98);
		sdp.setOptions(opcionesSdp);
		try {
			sdp.setIp(InetAddress.getLocalHost().getHostAddress().toString());
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		sdp.setPort(cliente.getPuertoUa());
		
		mensajeOk.setSdp(sdp);
		
		return enviarMensaje(mensajeOk);
		
	}
	private void responderACK(NotFoundMessage mensajeSip){
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias() );
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getFromName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("2 ACK\n");
		ack.setMaxForwards(UA.MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
	}
	
	private void responderACK(BusyHereMessage mensajeSip){
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias());
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getFromName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("2 ACK\n");
		ack.setMaxForwards(UA.MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
	}
	
	private void responderACK(OKMessage mensajeSip){
		ACKMessage ack=new ACKMessage();
		
		ack.setVias(mensajeSip.getVias() );
		ack.setFromUri(mensajeSip.getFromUri());
		ack.setToName(mensajeSip.getToName());
		ack.setFromName(mensajeSip.getFromName());
		ack.setDestination("sip:"+mensajeSip.getToName()+"@"+UA.dominio);
		ack.setToUri(mensajeSip.getToUri());
		ack.setCallId(mensajeSip.getCallId());
		ack.setcSeqNumber("2 ACK\n");
		ack.setMaxForwards(UA.MAXFORWARD);
		ack.setContentLength(0);
		
		enviarMensaje(ack);
	}
}
